/*
 Multi-MBC Loader
 Version: 1.0
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 25/08/2018
 Last Modified: 16/02/2019
 
 Drag and drop a single MBC1/3 ROM file to this executable and it will create a file name with "+iG_Loader" at the end
 for you to flash to the insideGadgets Gameboy 4MB 32KB/128KB SRAM Flash Cart.
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <libgen.h>

#ifdef _WIN32
#include <Windows.h>
#include <direct.h>
#define GetCurrentDir _getcwd
#else
#include <unistd.h>
#define GetCurrentDir getcwd
#endif

void delay_ms(uint16_t ms) {
	#if defined (_WIN32)
		Sleep(ms);
	#else
		struct timespec ts;
		ts.tv_sec = ms / 1000;
		ts.tv_nsec = (ms * 1000000) % 1000000000;
		nanosleep(&ts, NULL);
	#endif
}

// Read one letter from stdin
char read_one_letter (void) {
	char c = getchar();
	while (getchar() != '\n' && getchar() != EOF);
	return c;
}

int main(int argc, char **argv) {
	
	printf("insideGadgets - Multi-MBC Loader v1.0\n");
	printf("#############################################################\n");
	printf("Use this loader to boot MBC1/3 games.\nThis only works with the insideGadgets Gameboy 4MB 32KB/128KB SRAM Flash Cart.\n\n");
	
	uint8_t buffer[336] = {0x00};
	uint8_t writeBuffer[1] = {0x00};
	
	if (argc >= 2) {
		// Calculate file sizes
		long fileSizeTotal = 32768; // Loader size
		for (uint8_t g = 1; g < argc; g++) {
			FILE *romReadFile = fopen(argv[g], "rb");
			fseek (romReadFile, 0, SEEK_END);
			long fileSize = ftell (romReadFile);
			fileSizeTotal += fileSize;
			fclose(romReadFile);
		}
		
		// Check if too large
		if (fileSizeTotal > 4194304) {
			printf("ROM file too large, more than 4MB: %i\n", fileSizeTotal);
			read_one_letter();
			return 0;
		}
		
		// Check MBC type
		uint16_t startLocation = 0x0147;
		FILE *romCheckFile = fopen(argv[1], "rb");
		fseek(romCheckFile, startLocation, SEEK_SET);
		fread(buffer, 1, 1, romCheckFile);
		uint8_t mbcType = buffer[0];
		
		if (mbcType >= 0x19) {
			printf("The ROM supplied is an MBC5 ROM, you don't need to use the loader for this.\nPress any key.\n");
			fclose(romCheckFile);
			read_one_letter();
			return 1;
		}
		
		
		// Extract the filename before the dot
		char tempTitle[254];
		char outputFileName[200];
		char *dotLocation;
		strncpy(tempTitle, basename(argv[1]), 200);
		dotLocation = strrchr(tempTitle, '.'); // Check for the last dot in the file name
		
		if (dotLocation-tempTitle > 1) {
			strncpy(outputFileName, tempTitle, dotLocation-tempTitle);
			outputFileName[dotLocation-tempTitle] = 0;
		}
		else {
			strncpy(outputFileName, tempTitle, 18);
		}
		strncat(outputFileName, "+iG_Loader.gb", 14);
		//printf("%s\n", outputFileName);
		
		// Copy the base rom file to the output file
		char copyCommand[254];
		strncpy(copyCommand, "copy base.gb \"", 15);
		strncat(copyCommand, outputFileName, 200);
		strncat(copyCommand, "\" > /nul", 8);
		system(copyCommand);
		//printf("%s\n", copyCommand);
		
		
		// Open file created
		FILE *romWriteFile = fopen(outputFileName, "r+b");
		
		// Read 0x134 to 0x014F and write this to the output file
		startLocation = 0x134;
		FILE *romReadFile = fopen(argv[1], "rb");
		fseek(romReadFile, startLocation, SEEK_SET);
		fread(buffer, 1, 0x1C, romReadFile);
		
		// Change MBC type to 5 and re-calculate header checksum
		buffer[0x13] = 0x1B;
		uint8_t headerChecksum = 0;
		for (uint16_t x = 0; x < 0x19; x++) {
			headerChecksum = headerChecksum - buffer[x] - 1;
		}
		buffer[0x19] = headerChecksum;
		
		// Write
		fseek(romWriteFile, 0x134, SEEK_SET);
		fwrite(buffer, 0x1C, 1, romWriteFile);
		
		
		// Read MBC type
		startLocation = 0x0147;
		fseek(romReadFile, startLocation, SEEK_SET);
		fread(buffer, 1, 1, romReadFile);
		mbcType = buffer[0];
		
		// Read ROM size
		startLocation = 0x0148;
		fseek(romReadFile, startLocation, SEEK_SET);
		fread(buffer, 1, 1, romReadFile);
		uint8_t romSize = buffer[0];
		
		// Read RAM size
		startLocation = 0x0149;
		fseek(romReadFile, startLocation, SEEK_SET);
		fread(buffer, 1, 1, romReadFile);
		uint8_t ramSize = buffer[0];
		
		// Write the MBC type
		if (mbcType <= 0x0D && romSize <= 0x04 && ramSize <= 0x02) { // MBC1 <= 512KB ROM / 8KB RAM
			writeBuffer[0] = 0x2A;
		}
		else if (mbcType <= 0x0D && romSize <= 0x04 && ramSize == 0x03) { // MBC1 <= 512KB ROM / 32KB RAM
			writeBuffer[0] = 0x0A;
		}
		else if (mbcType <= 0x0D && romSize >= 0x05 && ramSize <= 0x02) { // MBC1 >= 1MB ROM / 8KB RAM
			writeBuffer[0] = 0x2C;
		}
		else if (mbcType <= 0x0D && romSize >= 0x05 && ramSize == 0x03) { // MBC1 >= 1MB ROM / 32KB RAM
			writeBuffer[0] = 0x0C;
		}
		else if (mbcType >= 0x0E && mbcType <= 0x18) { // MBC3
			writeBuffer[0] = 0x0E;
		}
		fseek(romWriteFile, 0x3000, SEEK_SET);
		fwrite(writeBuffer, 1, 1, romWriteFile);
		fclose(romWriteFile);
		fclose(romReadFile);
		
		char appendCommand[500];
		strncpy(appendCommand, "copy /b \"", 10);
		strncat(appendCommand, outputFileName, 100);
		strncat(appendCommand, "\"+\"", 4);
		strncat(appendCommand, argv[1], 200);
		strncat(appendCommand, "\" > /nul", 9);
		//printf("%s\n", appendCommand);
		
		system(appendCommand);
		
		printf("Completed\n");
		delay_ms(1000);
	}
	else {
		printf("You need to drag and drop your ROM files to this executable. Press any key.\n");
		read_one_letter();
	}
	
	return 0;
}